/*
ChrispyChars Dances
	by Lach

beta
*/

if not CC_DanceSkins
	rawset(_G, "CC_DanceSkins", {})
end

--[[
USAGE: Copy the above into your script.

Make a CC_DanceSkins entry for each skin you want to add a dance to.
This should consist of a table of tables, each of which is one dance.
When a player presses the dance button for the first time, the first dance is attempted.
Each successive press of the dance button increments the selected dance.
It is possible to only allow certain dances under certain conditions so that they are skipped!

These are the parameters each dance entry can have:
	- condition: function(player)
		If supplied, this function is run when the player is attempting to switch to this dance.
		If the function returns true, the dance is run, otherwise it is skipped.
		If a condition function is not supplied, the dance is always run.
	- keys: A table of further entries that control which sprites are displayed during a dance and for how long.
		You can also supply functions to be run when a key is switched into.
		If not supplied, the player stops dancing.
		More info on this below.

These are the parameters each entry in the keys table may have:
	- sprite2: The SPR2 set to use. FF_SPR2SUPER is automatically added to this if the player is super.
		Cannot be omitted.
	- frame: The frame of the SPR2 set to use.
		Cannot be omitted.
	- tics: The duration of this key, in tics. When the timer is up, the player moves on to the next key.
		Cannot be omitted.
	- angle: If supplied, the player's drawangle will be forced to this value, relative to the player's camera.
	- start: function(player)
		If supplied, this function is run when the player switches to this key.
	- think: function(player)
		If supplied, this function is run for each tic the player is in this key.
	- queuetransition: Boolean that determines whether the current key's duration must expire before the player switches to this key.

See the other scripts in the ChrispyChars addon for some examples!

Include an **UNEDITED** copy of this script with your custom characters and have it load before their scripts.
You can also choose not to include the script, in which case your character's dances will only be available when ChrispyChars or another CC_DANCES-using character is added.
Please note that not all features of this script have been extensively tested, as they were only built with ChrispyChars in mind.
If you find bugs from using a parameter in a way I have not accounted for, please post a report in ChrispyChars' SRB2 Message Board Releases thread!
]]--

if CC_DANCES
	return // script is already loaded in another addon!
end

rawset(_G, "CC_DANCES", true)

freeslot("S_CHRISPY_DANCE")

states[S_CHRISPY_DANCE] = {
	sprite = SPR_PLAY,
	tics = -1
}

local function PlayerCanDance(player)
	local mo = player.mo
	
	return not player.powers[pw_carry]
	and not player.powers[pw_nocontrol]
	and not (player.pflags & (PF_SLIDING|PF_SPINNING|PF_GLIDING|PF_BOUNCING|PF_STASIS|PF_STARTDASH))
	and not (player.cmd.forwardmove or player.cmd.sidemove)
	and not player.cc_nodance
end

local function ChangeDance(player)
	local skin = player.mo.skin
	local danceinfo = CC_DanceSkins[skin]
	local dance = player.cc_dance
	local currentnum = dance.num
	local num = currentnum % #danceinfo + 1
				
	while danceinfo[num].condition
	and num ~= dance.num
		local result = danceinfo[num].condition(player)
		
		if result == true
			break // do this dance!
		end
		
		// otherwise, continue looking
		num = $ % #danceinfo + 1
		
		// ...unless you looped back around
		if num == 1 and dance.num == 0
		or num == currentnum
			return false
		end
	end
	
	if not danceinfo[num].keys // no keys?! don't dance then D:
		return false
	end
	
	if dance.key > #danceinfo[num].keys
		dance.key = 1
	end
	if dance.music ~= danceinfo[num].music
		dance.music = danceinfo[num].music
		if dance.music == nil
			P_RestoreMusic(player)
		else
			S_ChangeMusic(dance.music, true, player)
		end
	end
	dance.num = num
	dance.skin = skin
	return true
end

local function QuitDance(player, dance)
	if dance.music
		P_RestoreMusic(player)
		dance.music = nil
	end
	dance.num = 0
end

addHook("ThinkFrame", do
	for player in players.iterate
		local mo = player.mo
		local danceinfo = mo and CC_DanceSkins[mo.skin]
		
		if not danceinfo
			local dance = player.cc_dance
			if dance
				if dance.music
					P_RestoreMusic(player)
				end
				player.cc_dance = nil
				if mo
					mo.cc_dance = nil
				end
			end
			continue
		end
		
		local dance = mo.cc_dance
		if not mo.cc_dance
			dance = {}
			mo.cc_dance = dance
			player.cc_dance = dance
			
			dance.buttontime = INT32_MIN
		end
		
		// count length of time button is held for
		local button = player.cc_buttonmap and player.cc_buttonmap.dance or BT_CUSTOM2
		
		if player.cmd.buttons & button
			dance.buttontime = $ + 1
		else
			dance.buttontime = 0
		end
		
		// pressed the button? change dance animation!
		if dance.buttontime == 1
			if mo.state == S_CHRISPY_DANCE
			and not dance.queuechange
				if danceinfo[dance.num].keys[dance.key].queuetransition
					dance.queuechange = true
				else
					ChangeDance(player)
				end
			elseif PlayerCanDance(player)
			and P_IsObjectOnGround(mo)
				if mo.neck and CC_KillNeck // sneaky Neck support
					CC_KillNeck(mo.neck)
				end
				dance.num = 0
				dance.key = 0
				dance.tics = 0
				dance.music = nil
				if ChangeDance(player)
					mo.state = S_CHRISPY_DANCE
					player.panim = PA_DASH
				end
			end
		end
		
		if mo.state == S_CHRISPY_DANCE
			if not PlayerCanDance(player)
				QuitDance(player, dance)
				mo.state = S_PLAY_WALK
			else
				if mo.skin ~= dance.skin // catch skin changes
					ChangeDance(player)
				end
				
				local keys = danceinfo[dance.num].keys
				local key
				
				// countdown
				dance.tics = $ - 1
				
				if dance.tics <= 0 // change keyframe when countdown is up
					if dance.queuechange
						ChangeDance(player)
						keys = danceinfo[dance.num].keys
						dance.queuechange = false
					end
					dance.key = $ % #keys + 1
					dance.tics = keys[dance.key].tics
					if keys[dance.key].start
						keys[dance.key].start(player)
					end
				end
				key = keys[dance.key]
				
				// set sprite and angle
				mo.sprite2 = key.sprite2 | ($ & FF_SPR2SUPER)
				mo.frame = key.frame | ($ & ~FF_FRAMEMASK)
				if key.angle ~= nil
					player.drawangle = (player.cmd.angleturn << 16) + key.angle
				end
				
				// run function
				if key.think
					key.think(player)
				end
			end
		elseif dance.num
			QuitDance(player, dance)
		end
	end
end)