/*
ChrispyChars Tricks
	by Lach

beta
*/

if not CC_TrickSkins
	rawset(_G, "CC_TrickSkins", {})
end

--[[
USAGE: Copy the above into your script.

Make a CC_TrickSkins entry for each skin you want to add extra tricks to.
The skin entry must have a table of trick entries.
It can also optionally enable the following parameters:
	- alwaysenabled: lets the character perform tricks regardless of whether the command is toggled
	- bouncetricks: enables enemy/monitor bounce tricks
Each trick entry in the skin's table can have multiple trick entries.
The available trick entry parameters are:
	- rarity: fixed-scale multiplier for how often the trick can be picked
		(FRACUNIT is default, 2*FRACUNIT would be twice as likely)
	- sprite2: SPR2 set the trick pose should use
	- rotation: direction that determines which way the trick should spin
		(0 for no rotation, > 0 for counterclockwise, < 0 for clockwise, "random" for either 1/-1)
	- angle: angle, relative to the camera, that the trick will be started from;
		needs to be an angle value, see https://wiki.srb2.org/wiki/Constants#Angles_.28ANG.2A.2FANGLE_.2A.29
		the player will return to their previous angle after the trick has finished (unless it rotates)
	- tics: if the trick should be animated, each frame will display for this long
		(if the trick should not be animated, set to 0 or below)
	- frames: if the trick is animated, this is the number of frames in the SPR2 animation
	- frame: if the trick is not animated, this is the frame of the SPR2 set to display

See the other scripts in the ChrispyChars addon for some examples!
When ChrispyChars is loaded, your tricks will take effect when the trick command is toggled on.
You can also copy this script and its sounds into your addon if you want the tricks to take effect regardless!
Just please don't edit it. :)
]]--

CC_TrickSkins["DEFAULT"] = {
	alwaysenabled = false,
	{
		sprite2 = SPR2_SPNG,
		rotation = "random",
		tics = 0,
	},
}

-- No need to copy anything below here! (and please don't edit it!)

if CC_TRICKS
	return
end

// To prevent duplicate freeslots
local function SafeFreeslot(...)
	for _, item in ipairs({...})
		if rawget(_G, item) == nil
			freeslot(item)
		end
	end
end

SafeFreeslot("SPR_CONF", "S_CHRISPY_CONFETTI", "S_CHRISPY_TRICK", "MT_CCTRICKSCORE", "MT_CHRISPY_CONFETTI")

states[S_CHRISPY_CONFETTI] = {
	sprite = SPR_CONF,
	frame = FF_PAPERSPRITE,
	action = A_RollAngle,
	var1 = 10,
	tics = 1,
	nextstate = S_CHRISPY_CONFETTI,
}

states[S_CHRISPY_TRICK] = {
	sprite = SPR_PLAY,
	frame = SPR2_SPNG,
	tics = 14,
	var1 = SPR2_EDGE | (SPR2_GASP << 8),
	nextstate = S_PLAY_SPRING,
}

mobjinfo[MT_CCTRICKSCORE] = {
	spawnstate = S_SCRA,
	height = mobjinfo[MT_SCORE].height,
	radius = mobjinfo[MT_SCORE].radius,
	speed = mobjinfo[MT_SCORE].speed,
	flags = mobjinfo[MT_SCORE].flags | MF_NOCLIPHEIGHT,
}

mobjinfo[MT_CHRISPY_CONFETTI] = {
	spawnstate = S_CHRISPY_CONFETTI,
	radius = 2*FRACUNIT,
	height = 4*FRACUNIT,
	flags = MF_SCENERY|MF_NOCLIPTHING,
	painchance = ANG15, // rotation speed
	speed = 2*FRACUNIT, // max fall speed
	mass = 3*FRACUNIT/4, // momentum reduction rate
	reactiontime = 6*TICRATE, // lifetime
}

SafeFreeslot("sfx_trickf")
sfxinfo[sfx_trickf].caption = "/"
for i = 1, 6
	SafeFreeslot("sfx_trick"..i)
	sfxinfo[sfx_trick1 + i - 1].caption = "/"
end

rawset(_G, "CC_TRICKS",
{
	{
		text = "GOOD!",
		color = SKINCOLOR_SAPPHIRE,
		sound = sfx_trick1,
		score = 10,
		state = S_SCRL,
	},
	{
		text = "GREAT!",
		color = SKINCOLOR_SKY,
		sound = sfx_trick1,
		score = 10,
		state = S_SCRL,
	},
	{
		text = "NICE!",
		color = SKINCOLOR_AQUA,
		sound = sfx_trick2,
		score = 100,
		state = S_SCRA,
	},
	{
		text = "JAMMIN'!",
		color = SKINCOLOR_PERIDOT,
		sound = sfx_trick2,
		score = 100,
		state = S_SCRA,
	},
	{
		text = "COOL!",
		color = SKINCOLOR_YELLOW,
		sound = sfx_trick3,
		score = 200,
		state = S_SCRB,
	},
	{
		text = "RADICAL!",
		color = SKINCOLOR_SUNSET,
		sound = sfx_trick3,
		score = 200,
		state = S_SCRB,
	},
	{
		text = "TIGHT!",
		color = SKINCOLOR_ORANGE,
		sound = sfx_trick4,
		score = 400,
		state = S_SCRF,
	},
	{
		text = "AWESOME!",
		color = SKINCOLOR_FLAME,
		sound = sfx_trick4,
		score = 500,
		state = S_SCRC,
	},
	{
		text = "EXTREME!",
		color = SKINCOLOR_MAGENTA,
		sound = sfx_trick5,
		score = 800,
		state = S_SCRG,
	},
	{
		text = "PERFECT!",
		color = -1,
		sound = sfx_trick6,
		score = 1000,
		state = S_SCRD,
	},
})

local function valid(mo)
	return mo and mo.valid
end

rawset(_G, "CC_MIN_HIGHSCORE", #CC_TRICKS)

local TF_CANTRICK = 1 << 0
local TF_NOSCORE = 1 << 1

// Activation

local function PlayerCanTrick(player)
	local mo = player.mo
	
	return not P_IsObjectOnGround(mo)
	and P_MobjFlip(mo)*mo.momz > 0
	and not (player.pflags & (PF_JUMPED|PF_SPINNING|PF_GLIDING|PF_BOUNCING|PF_SLIDING))
	and player.powers[pw_carry] == CR_NONE
	and not player.powers[pw_tailsfly]
	and not player.cantrick // xmom
end

local function PlayTrickSound(mo, count)
	if count < 1
		if count == -1
			S_StartSound(mo, sfx_trickf)
		end
		return
	end
	
	count = min($, #CC_TRICKS)
	S_StartSound(mo, CC_TRICKS[count].sound)
end

rawset(_G, "CC_DEFAULT_CONFETTI_COLORS", { // This default list is changed in CHRISPYCHARS.lua, but it's set up with vanilla colors here so this script can be used in other addons
	SKINCOLOR_NONE,
	SKINCOLOR_NONE,
	SKINCOLOR_AETHER,
	SKINCOLOR_KETCHUP,
	SKINCOLOR_YELLOW,
	SKINCOLOR_EMERALD,
	SKINCOLOR_ICY,
	SKINCOLOR_BUBBLEGUM,
	SKINCOLOR_NEON
})

rawset(_G, "CC_SpawnConfetti", function(player, num, speed, color, docenter, func)
	local mo = player.mo
	local grounded = P_IsObjectOnGround(mo) or player.powers[pw_carry] ~= CR_NONE
	speed = $ >> FRACBITS
	color = $ or 0
	for i = 1, num
		local part = P_SpawnMobjFromMobj(mo, 0, 0, player.height, MT_CHRISPY_CONFETTI)
		part.target = mo
		if speed
			part.momx = P_RandomRange(-speed, speed)*mo.scale
			part.momy = P_RandomRange(-speed, speed)*mo.scale
			P_SetObjectMomZ(part, P_RandomRange(not grounded and -speed or 0, speed)*FRACUNIT, false)
			if not docenter
				P_TeleportMove(part,
					part.x + part.momx,
					part.y + part.momy,
					part.z + part.momz)
			end
			part.momz = $ + mo.pmomz
		end
		if color > 0
			part.color = color
		else
			local colorlist = player.cc_confetticolors or CC_DEFAULT_CONFETTI_COLORS
			part.color = colorlist[P_RandomKey(#colorlist) + 1] or player.skincolor
		end
		if func
			func(part)
		end
	end
end)

rawset(_G, "CC_SetTrickState", function(mo)
	local tricks = mo.cc_tricks
	local trickset = CC_TrickSkins[mo.skin] or CC_TrickSkins["DEFAULT"]
	local selection = 0
	local chance = 0
	local totalchance = 0
	local player = mo.player
	
	if not tricks then return end
	
	for i = 1, #trickset
		totalchance = $ + (trickset[i].rarity or FRACUNIT)
	end
	
	totalchance = FixedMul(P_RandomFixed(), $)
	
	while chance <= totalchance
		selection = $ + 1
		chance = $ + (trickset[selection].rarity or FRACUNIT)
	end
	
	trickset = $[selection]
	
	mo.state = S_CHRISPY_TRICK
	mo.sprite2 = trickset.sprite2 | (player.powers[pw_super] and FF_SPR2SUPER)
	if trickset.frame ~= nil
		mo.frame = trickset.frame | ($ & ~FF_FRAMEMASK)
	end
	tricks.frames = trickset.frames
	tricks.duration = trickset.tics
	tricks.tics = tricks.duration
	tricks.rotation = trickset.rotation
	if tricks.rotation == "random"
		tricks.rotation = -1 + 2*P_RandomKey(2)
	end
	if trickset.angle ~= nil
		if tricks.angle == "random"
			tricks.angle = FixedAngle(P_RandomKey(360)*FRACUNIT)
		else
			tricks.angle = (player.cmd.angleturn << 16) + trickset.angle
		end
	elseif tricks.rotation
		tricks.angle = mo.player.drawangle
	else
		tricks.angle = nil
	end
	tricks.count = max($, 0)
end)

rawset(_G, "CC_DoChrispyTrick", function(mo, force)
	local tricks = mo.cc_tricks
	local player = mo.player
	
	if not tricks
	or P_IsObjectOnGround(mo)
		return
	end
	
	if force
	or not (tricks.flags & TF_NOSCORE)
		tricks.count = $ + 1
		if tricks.count < 1
			return
		end
		
		local trick = CC_TRICKS[min(tricks.count, #CC_TRICKS)]
		PlayTrickSound(mo, tricks.count)
		CC_SpawnConfetti(player, 12, 12*FRACUNIT, trick.color, force)
		if not force
		and gametyperules & GTR_FRIENDLY
			local obj = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_CCTRICKSCORE)
			obj.target = mo
			obj.state = trick.state
			obj.color = max(trick.color, 0)
			obj.colorized = obj.color and true or false
			P_AddPlayerScore(player, trick.score)
		end
	else
		PlayTrickSound(mo, -1)
	end
end)

addHook("JumpSpecial", function(player)
	local mo = player.mo
	local trickset = CC_TrickSkins[mo.skin] or CC_TrickSkins["DEFAULT"]
	local tricks = mo.cc_tricks
	
	if not (player.pflags & (PF_JUMPDOWN))
	and (player.cc_tricksenabled or trickset.alwaysenabled)
	and not (player.powers[pw_super] and trickset.nosupertricks)
	and tricks and tricks.flags & TF_CANTRICK
	and PlayerCanTrick(player)
		if mo.state ~= S_CHRISPY_TRICK
			CC_SetTrickState(mo)
			CC_DoChrispyTrick(mo)
		end
		return true
	end
end)

local function AddTrickCollider(mo, obj)
	if not (mo.valid and obj.valid)
	or mo.eflags & MFE_SPRUNG
		return
	end
	
	local player = mo.player
	local tricks = mo.cc_tricks
	local trickset = CC_TrickSkins[mo.skin] or CC_TrickSkins["DEFAULT"]
	local moz = mo.z
	if obj.flags & MF_ENEMY // for some reason you don't... actually collide with spring shells :V
		moz = $ + mo.momz
	end
	
	if not tricks
	or not (player.cc_tricksenabled or trickset.alwaysenabled)
	or not (moz <= obj.z + obj.height and obj.z <= moz + mo.height)
		return
	end
	
	for _, col in ipairs(tricks.colliders)
		if col == obj
			return
		end
	end
	table.insert(tricks.colliders, obj)
end

addHook("MobjMoveCollide", AddTrickCollider, MT_PLAYER)
addHook("MobjCollide", AddTrickCollider, MT_PLAYER)

addHook("ThinkFrame", do
	for player in players.iterate
		local mo = player.mo
		if not mo
			continue
		end
		
		local trickset = CC_TrickSkins[mo.skin] or CC_TrickSkins["DEFAULT"]
		
		if not (player.cc_tricksenabled or trickset.alwaysenabled)
			if mo.cc_tricks
				mo.cc_tricks = nil
			end
			continue
		end
		
		local grounded = P_IsObjectOnGround(mo)
		
		if not mo.cc_tricks
			local tricks = {}
			tricks.count = -1
			tricks.angle = 0
			tricks.grounded = grounded
			tricks.flags = 0
			tricks.blink = 0
			tricks.springs = {}
			tricks.colliders = {}
			mo.cc_tricks = tricks
			
			if player.cc_tricks_highscore == nil
				player.cc_tricks_highscore = CC_MIN_HIGHSCORE
			end
		end
		
		local tricks = mo.cc_tricks
		
		if (mo.eflags & MFE_SPRUNG)
			tricks.flags = $ & ~TF_NOSCORE | TF_CANTRICK
			for _, obj in ipairs(tricks.colliders)
				for _, spring in ipairs(tricks.springs)
					if obj == spring
						tricks.flags = $ | TF_NOSCORE
						break
					end
				end
				
				if tricks.flags & TF_NOSCORE
					break
				end
				table.insert(tricks.springs, obj)
			end
		end
		
		if #tricks.colliders
			tricks.colliders = {}
		end
		
		if mo.state == S_CHRISPY_TRICK
			if tricks.rotation
				player.drawangle = tricks.angle + tricks.rotation*FixedAngle((mo.tics - 1)*360*FRACUNIT/states[mo.state].tics)
			elseif tricks.angle ~= nil
				player.drawangle = tricks.angle
			end
			
			if tricks.duration > 0
				tricks.tics = $ - 1
				if tricks.tics <= 0
					mo.frame = ((($ + 1) & FF_FRAMEMASK) % tricks.frames) | ($ & ~FF_FRAMEMASK)
					tricks.tics = tricks.duration
				end
			end
		end
		
		if tricks.flags & TF_CANTRICK
		and not PlayerCanTrick(player)
			tricks.flags = $ & ~TF_CANTRICK
		end
		
		if tricks.blink > 0
			tricks.blink = $ - 1
		end
		
		if grounded and not tricks.grounded
			if tricks.count > player.cc_tricks_highscore
				player.cc_tricks_highscore = tricks.count
				tricks.blink = 2*TICRATE
				S_StartSound(nil, sfx_chchng, player)
			end
			tricks.count = -1
			tricks.springs = {}
			tricks.flags = 0
		end
		
		tricks.grounded = grounded
	end
end)

// HUD

hud.add(function(v, player)
	local mo = player.mo
	if not mo
		return
	end
	
	local tricks = mo.cc_tricks
	
	if not tricks return end
	
	local count = tricks.count
	local x = 16*FRACUNIT
	local y = 60*FRACUNIT
	local flags = V_SNAPTOTOP | V_SNAPTOLEFT | V_HUDTRANS | V_PERPLAYER
	local scale = FRACUNIT >> 1
	
	// high score
	if count > player.cc_tricks_highscore
		local scale = scale >> 1
		local x = x
		local firsthypercolor = SKINCOLOR_SUPERSILVER3 + 5
		local hypercolors = (SKINCOLOR_SUPERPURPLE3 + 5 - firsthypercolor) / 5
		local text = "NEW  RECORD!"
		
		for i = 1, text:len()
			local n = player.jointime + i
			local y = y + 16*FRACUNIT + 2*sin(n*ANG20)
			local letter = text:sub(i, i)
			local color = firsthypercolor + (n % hypercolors) * 5
			v.drawScaledNameTag(x, y, letter, flags, scale, color, SKINCOLOR_BONE)
			x = $ + (v.nameTagWidth(letter) + 1)*scale
		end
	elseif count > 0 or (tricks.blink >> 1) & 1
		local scale = scale >> 1
		v.drawScaledNameTag(x, y + 16*FRACUNIT, "HIGHSCORE  #"..player.cc_tricks_highscore, flags, scale, SKINCOLOR_SILVER, SKINCOLOR_BONE)
	end
	
	if count < 1
		return
	end
	
	local trick = CC_TRICKS[min(count, #CC_TRICKS)]
	local text = trick.text
	
	// current score
	if trick.color < 0 // animated
		local length = text:len()
		
		text = $ .. "  #" .. tricks.count
		
		for i = 1, text:len()
			local n = player.jointime + i
			local y = y + 3*sin(n*ANG20)
			local letter = text:sub(i, i)
			local color
			if i > length
				color = SKINCOLOR_SILVER
			else
				color = SKINCOLOR_RUBY + n % (FIRSTSUPERCOLOR - SKINCOLOR_RUBY)
			end
			v.drawScaledNameTag(x, y, letter, flags, scale, color, SKINCOLOR_BONE)
			x = $ + (v.nameTagWidth(letter) + 1)*scale
		end
	else
		local n = player.jointime
		y = 60*FRACUNIT + 3*sin(n*ANG20)
		v.drawScaledNameTag(x, y, text, flags, scale, trick.color, SKINCOLOR_BONE)
		x = $ + v.nameTagWidth(text)*scale
		v.drawScaledNameTag(x, y, "  #"..tricks.count, flags, scale, SKINCOLOR_SILVER, SKINCOLOR_BONE)
	end
end)

// Bounce chain

local function BounceTrick(item, mo, itemflags)
	if not (item.valid and item.flags & itemflags)
	or not (mo and mo.valid and mo.player)
		return
	end
	
	local player = mo.player
	local trickset = CC_TrickSkins[mo.skin] or CC_TrickSkins["DEFAULT"]
	
	if not trickset.bouncetricks
	or not (trickset.alwaysenabled or player.cc_tricksenabled)
	or player.powers[pw_super] and trickset.nosupertricks
		return
	end
	CC_DoChrispyTrick(mo, true)
end

addHook("MobjDamage", function(item, mo)
	BounceTrick(item, mo, MF_ENEMY|MF_BOSS)
end)

addHook("MobjDeath", function(item, mo)
	BounceTrick(item, mo, MF_MONITOR)
end)

// Object thinkers

addHook("MobjThinker", function(mo)
	local target = mo.target
	
	if not valid(target) return end
	
	local z
	if target.eflags & MFE_VERTICALFLIP
		mo.eflags = $ | MFE_VERTICALFLIP
		z = target.z
	else
		mo.eflags = $ & ~MFE_VERTICALFLIP
		z = target.z + target.height
	end
	
	z = $ + P_MobjFlip(mo) * (states[mo.state].tics - mo.tics) * FixedMul(mo.info.speed, mo.scale)
	P_TeleportMove(mo, target.x, target.y, z)
end, MT_CCTRICKSCORE)

addHook("MobjSpawn", function(mo)
	mo.movedir = FixedAngle(360*P_RandomFixed())
	mo.angle = mo.movedir
	mo.fuse = mo.info.reactiontime
end, MT_CHRISPY_CONFETTI)

addHook("MobjThinker", function(mo)
	// kill when hitting the floor
	if P_IsObjectOnGround(mo)
		P_RemoveMobj(mo)
		return
	end
	
	// reduce speed
	if not (mo.flags2 & MF2_DONTRESPAWN)
		mo.momx = FixedMul($, mo.info.mass)
		mo.momy = FixedMul($, mo.info.mass)
		mo.momz = FixedMul($, mo.info.mass)
		if FixedHypot(FixedHypot(mo.momx, mo.momy), mo.momz) < mo.scale
			mo.momx = 0
			mo.momy = 0
			mo.flags2 = $ | MF2_DONTRESPAWN
		end
	elseif not (mo.flags & MF_NOGRAVITY) // limit fall speed
		local flip = P_MobjFlip(mo)
		local maxfall = -FixedMul(mo.info.speed, mo.scale)
		if flip*mo.momz < maxfall
			mo.momz = flip*FixedMul(flip*$, mo.info.mass)
			if flip*mo.momz > maxfall
				mo.momz = flip*maxfall
				mo.flags = $ | MF_NOGRAVITY
			end
		end
	end
	
	// rotate
	mo.angle = $ + mo.info.painchance
end, MT_CHRISPY_CONFETTI)

// Commands

local function Boolean(arg)
	local num = tonumber(arg)
	if num ~= nil
		return num > 0
	end
	
	if arg == "on"
	or arg == "yes"
	or arg == "true"
		return true
	elseif arg == "off"
	or arg == "no"
	or arg == "false"
		return false
	end
end

local function GetEnabledString(val)
	return val and "enabled." or "disabled."
end

local function ToggleTricks(player, arg)
	local val
	if arg
		arg = $:lower()
		val = Boolean(arg)
		if val ~= nil
			player.cc_tricksenabled = val
			if not player.cc_nosave
				CONS_Printf(splitscreen and consoleplayer or player, "Tricks "..GetEnabledString(player.cc_tricksenabled))
				if CC_SaveConfiguration
					CC_SaveConfiguration(player)
				end
			end
			return
		end
	end
	
	CONS_Printf(splitscreen and consoleplayer or player, "Enables/disables tricks usable by all characters.")
end

COM_AddCommand("tricks2", ToggleTricks, 2)
COM_AddCommand("tricks", ToggleTricks)

local function GenerateConfettiList(player)
	local colorlist
	if not player.cc_confetticolors
		colorlist = "\n\t- Default"
	else
		colorlist = ""
		for _, color in ipairs(player.cc_confetticolors)
			colorlist = $ .. "\n\t- " .. (color and R_GetNameByColor(color) or "None")
		end
	end
	
	return colorlist
end

local function SetPlayerConfetti(player, ...)
	local args = {...}
	local result
	local setcolors = {}
	local addcolors = {}
	local removecolors = {}
	
	if not #args
		local desc = "Sets up a custom list of colors to be randomly selected from when spawning confetti." +
			"\nEnter any number of colors to set your list." +
			"\nUse + or - prefixes to add colors to or subtract colors from your existing custom list." +
			"\nUse \"none\" to substitute your player color." +
			"\nUse \"default\" as the only entry to delete your custom list." +
			"\nCurrent color list:" +
			GenerateConfettiList(player)
		
		CONS_Printf(splitscreen and consoleplayer or player, desc)
		return
	end
	
	if #args == 1 and args[1]:lower() == "default"
		player.cc_confetticolors = nil
	else
		for _, item in ipairs(args)
			local firstchar = item:sub(1, 1)
			local arr = setcolors
			
			if firstchar == "+"
				item = item:sub(2)
				arr = addcolors
			elseif firstchar == "-"
				item = item:sub(2)
				arr = removecolors
			end
			
			local color = tonumber(item)
			
			if color ~= nil
				if not (color >= 0 and color < #skincolors)
					CONS_Printf(splitscreen and consoleplayer or player, "Skincolor "..n.." out of range (1 - "..(#skincolors-1)..").")
					return
				end
			else
				local constant = "SKINCOLOR_"..item:upper()
				local isvalid
				isvalid, color = pcall(function() return _G[constant] end)
				if not isvalid
					CONS_Printf(splitscreen and consoleplayer or player, constant .. " is not a valid color.")
					return
				end
			end
			table.insert(arr, color)
		end
		
		result = #setcolors and setcolors or player.cc_confetticolors or {}
		
		for _, item in ipairs(addcolors)
			table.insert(result, item)
		end
		
		for _, item in ipairs(removecolors)
			for i = #result, 1, -1
				if result[i] == item
					table.remove(result, i)
					break
				end
			end
		end
		
		if #result
			table.sort(result)
			player.cc_confetticolors = result
		else
			player.cc_confetticolors = nil
		end
	end
	
	if not player.cc_nosave
		local desc = "Confetti color list:" +
			GenerateConfettiList(player)
		
		CONS_Printf(splitscreen and consoleplayer or player, desc)
		if CC_SaveConfiguration
			CC_SaveConfiguration(player)
		end
	end
end

COM_AddCommand("confetti2", SetPlayerConfetti, 2)
COM_AddCommand("confetti", SetPlayerConfetti)
